/*
 * 	usart_buffered.c
 *  Buffered USART driver
 * Buffer handling need disabling/enabling interrupts...
 */
void startup(void) __attribute__((naked)) __attribute__((section (".start_section")) );

void startup ( void )
{
__asm volatile(
	" LDR R0,=0x2001C000\n"		/* set stack */
	" MOV SP,R0\n"
	" BL main\n"				/* call main */
	" B .\n"					/* never return */
	) ;
}

void EnableInterrupts( void ) __attribute__( ( naked ) ) ;
void EnableInterrupts( void )
{
	__asm (
	"		cpsie 	i\n"		/* set I=0 */
	"		bx		lr\n"
	);
}
void DisableInterrupts( void ) __attribute__( ( naked ) ) ;
void DisableInterrupts( void )
{
	__asm (
	"		cpsid 	i\n"		/* set I=1 */
	"		bx		lr\n"
	);
}

/* Buffering ... */
#define		USARTBUFFSIZE 32
typedef struct{
    unsigned char in;
    unsigned char out;
    unsigned char count;
    unsigned char buff[USARTBUFFSIZE];
}FIFO;

FIFO	RxBuf, TxBuf;


static void bufferInit( FIFO *buffer )
{
	buffer->count = 0;		//0 bytes in buffer
	buffer->in = 0;			//index points to start
	buffer->out = 0;		//index points to start
}

/* Kan också ge en introduktion till  "ömsesedig uteslutning"
 * ställ frågan, vad händer exempelvis om
 * bufferPut( Tx ) blir avbruten av bufferGet( Tx )
 * observera att variabeln 'count' delas av de båda funktionerna...
 */
static int bufferPut( FIFO *buffer, char ch )
{
	if( buffer->count==USARTBUFFSIZE ){
		return 0;		//buffer full
	}	
	DisableInterrupts();
	buffer->buff[buffer->in++]=ch;
	buffer->count++;
	if(buffer->in==USARTBUFFSIZE)
		buffer->in=0;  //wrap
	EnableInterrupts();
	return 1;
}

static int bufferGet( FIFO *buffer, char *ch)
{
	if(buffer->count==0)
		return 0;	//buffer empty
	DisableInterrupts();
	*ch=buffer->buff[buffer->out++];
	buffer->count--;
	if(buffer->out==USARTBUFFSIZE)
		buffer->out=0;//start from beginning
	EnableInterrupts();
	return 1;
}

/* USART definitions */

typedef struct tag_usart
{
	volatile unsigned short sr;
	volatile unsigned short Unused0;
	volatile unsigned short dr;
	volatile unsigned short Unused1;
	volatile unsigned short brr;
	volatile unsigned short Unused2;
	volatile unsigned short cr1;
	volatile unsigned short Unused3;
	volatile unsigned short cr2;
	volatile unsigned short Unused4;
	volatile unsigned short cr3;
	volatile unsigned short Unused5;
	volatile unsigned short gtpr;
} USART;

#define	EN		(1<<13)
#define	TXEIE	(1<<7)
#define	TCIE	(1<<6)
#define	RXNEIE	(1<<5)
#define	TE		(1<<3)
#define	RE		(1<<2)

#define	TXE		(1<<7)	/* Transmit data register empty */
#define	RXNE	(1<<5)	/* Receive data register full */

#define USART1  ((USART *) 0x40011000)

#define	USART1_IRQVEC 	((void (**)(void) ) 0x2001C0D4 ) 

typedef	void (*irq_routine )(void) ; 


#define	NVIC_USART1_ISER 		0xE000E104  
#define NVIC_USART1_IRQ_BPOS  	(1<<5)

void	spin(void)
{
	while(1);
}

static void	usart_service_routine( void )
{
	char	c;
	if( USART1->sr & RXNE)
	{ 
		if ( ! bufferPut( &RxBuf , (char) USART1->dr ) )
		{  spin();/* Buffer full ... */ }
	}		

 	if( (USART1->sr & TXE ) && ( USART1->cr1 & TXEIE ) )
	{
		if( bufferGet ( &TxBuf , &c ) )
		{
			USART1->dr = (unsigned short) c;
		}else{
			USART1->cr1 &= ~TXEIE;	/* Clear interrupt on transmitter empty */								
		}
	}
}

void	usart_init( void )
{
	*((unsigned int *) NVIC_USART1_ISER) |= NVIC_USART1_IRQ_BPOS;
	*((void (**)(void) ) 0x2001C0D4 )  = usart_service_routine;	
	/* Init buffers */
	bufferInit( &RxBuf );	
	bufferInit( &TxBuf );	
	
	/* Setup USART, receive with IRQ, polled transmit, 10 bit (1 startbit, 8 databits */ 
	USART1->brr = 0x2D9;	/* Baudrade = 115200 */
	USART1->cr3 = 0;
	USART1->cr2 = 0;	/* 1 stopbit */
	USART1->cr1 = EN | RXNEIE | TE | RE;
}




static char	usart_get ( void )
{
	char	c;
	if( bufferGet ( &RxBuf , &c ) )
		return c;
	return 0;
}

static void usart_put( char c )
{
	if( bufferPut ( &TxBuf , c ) )
	{
		if( ( USART1->cr1 & TXEIE )==0)
			USART1->cr1 |= TXEIE;	/* Interrupt on transmitter empty */
	}else{
		spin(); /* Buffer full... */
	}
}


/* 
 * print character string to console 
 */
void	printlusart(char *s)
{
	while(*s)
		usart_put(*s++);
}

/*
 * usart receive/transmit interrupt test
 */
 
void main(void)
{	
	char	c;
	usart_init();
	printlusart("USART program");
	while( 1 )
	{
		c = usart_get();
		if ( c )
			usart_put(c);
	}
}
