/*
 * 	usart_driver.c
 *  General interrupt driven buffered USART driver
 */
void startup(void) __attribute__((naked)) __attribute__((section (".start_section")) );

void startup ( void )
{
asm volatile(
	" LDR R0,=0x2001C000\n"		/* set stack */
	" MOV SP,R0\n"
	" BL main\n"				/* call main */
	".L1: B .L1\n"				/* never return */
	) ;
}


/* Buffering ... */
#define		USARTBUFFSIZE 4
typedef struct{
    unsigned char in;
    unsigned char out;
    unsigned char count;
    unsigned char buff[USARTBUFFSIZE];
}FIFO_TypeDef;

static void bufferInit( FIFO_TypeDef *buffer )
{
	buffer->count = 0;		//0 bytes in buffer
	buffer->in = 0;			//index points to start
	buffer->out = 0;		//index points to start
}

static int bufferPut( FIFO_TypeDef *buffer, char ch )
{
	if( buffer->count==USARTBUFFSIZE ){
		return 0;		//buffer full
	}	
	buffer->buff[buffer->in++]=ch;
	buffer->count++;
	if(buffer->in==USARTBUFFSIZE)
		buffer->in=0;  //wrap
	return 1;
}

static int bufferGet( FIFO_TypeDef *buffer, char *ch)
{
	if(buffer->count==0)
		return 0;	//buffer empty
	*ch=buffer->buff[buffer->out++];
	buffer->count--;
	if(buffer->out==USARTBUFFSIZE)
		buffer->out=0;//start from beginning
	return 1;
}

/* USART definitions */

typedef struct tag_usart
{
	volatile unsigned short sr;
	volatile unsigned short Unused0;
	volatile unsigned short dr;
	volatile unsigned short Unused1;
	volatile unsigned short brr;
	volatile unsigned short Unused2;
	volatile unsigned short cr1;
	volatile unsigned short Unused3;
	volatile unsigned short cr2;
	volatile unsigned short Unused4;
	volatile unsigned short cr3;
	volatile unsigned short Unused5;
	volatile unsigned short gtpr;
} USART;

#define	EN		(1<<13)
#define	TXEIE	(1<<7)
#define	TCIE	(1<<6)
#define	RXNEIE	(1<<5)
#define	TE		(1<<3)
#define	RE		(1<<2)

#define	TXE		(1<<7)	/* Transmit data register empty */
#define	RXNE	(1<<5)	/* Receive data register full */

typedef	void (*irq_routine )(void) ; 


#define	NVIC_USART1_ISER 		0xE000E104  
#define NVIC_USART1_IRQ_BPOS  	(1<<5)

#define	NVIC_USART2_ISER 		0xE000E104  
#define NVIC_USART2_IRQ_BPOS  	(1<<6)

/* Container for USART driver */
typedef struct{
	USART * 		usart;
	FIFO_TypeDef	TxBuf;
	FIFO_TypeDef	RxBuf;
	irq_routine		service; 
	unsigned int	usart_irq_vector; 	
} USART_DRIVER;

USART_DRIVER	drivers[2];	

void	usart_driver_init( USART_DRIVER * driver, USART * u, irq_routine s, unsigned int irq_vector)
{
	driver->usart = u;
	driver->service = s;
	if( irq_vector == 37 )
	{
		*((unsigned int *) NVIC_USART1_ISER) |= NVIC_USART1_IRQ_BPOS;
		*((void (**)(void) ) 0x2001C0D4 )  = s;	
	}
	if( irq_vector == 38 )
	{
		*((unsigned int *) NVIC_USART2_ISER) |= NVIC_USART2_IRQ_BPOS;
		*((void (**)(void) ) 0x2001C0D8 )  = s;	
	}
	/* Init buffers */
	bufferInit( &(driver->RxBuf) );	
	bufferInit( &(driver->TxBuf) );	
	
	/* Setup USART, receive with IRQ, polled transmit, 10 bit (1 startbit, 8 databits */ 
	u->brr = 0x2D9;	/* Baudrade = 115200 */
	u->cr3 = 0;
	u->cr2 = 0;	/* 1 stopbit */
	u->cr1 = EN | RXNEIE | TE | RE;
}


void spin(void)
{ 
		while(1);
}		





static void	usart_service_routine( USART_DRIVER *ud )
{
	char	c;
	USART *u = ud->usart;
	if( u->sr & RXNE)
	{ 
		if ( ! bufferPut( &(ud->RxBuf) , (char) u->dr ) )
		{  spin();/* Buffer full ... */ }
	}		

 	if( u->cr1 & TXEIE )
	{
		if( bufferGet ( &(ud->TxBuf) , &c ) )
		{
			if( u->sr & TXE )
				u->dr = (unsigned short) c;
		}else{
			u->cr1 &= ~TXEIE;	/* Clear interrupt on transmitter empty */								
		}
	}
}

static void	usart1_service_routine( void )
{
	usart_service_routine( &drivers[0] );
}

static void	usart2_service_routine( void )
{
	usart_service_routine( &drivers[1] );
}

static char	get ( USART_DRIVER *ud )
{
	char	c;
	if( bufferGet ( &(ud->RxBuf) , &c ) )
		return c;
	return 0;
}

static void put( USART_DRIVER *ud, char c )
{
	USART *u = ud->usart;
	if( bufferPut ( &(ud->TxBuf) , c ) )
	{
		if( ( u->cr1 & TXEIE )==0)
			u->cr1 |= TXEIE;	/* Interrupt on transmitter empty */
	}else{
		spin(); /* Buffer full... */
	}
}


#define	Console1 	&drivers[0]
#define	Console2 	&drivers[1]


void main(void)
{
	char	c;
	usart_driver_init( &drivers[0], (USART *) 0x40011000, usart1_service_routine , 37);
	usart_driver_init( &drivers[1], (USART *) 0x40004400, usart2_service_routine , 38);
	while( 1 )
	{
		if( c = get(Console1) )
			put( Console2, c );
		if( c = get( Console2 ))
			put( Console1, c );
	}
}

