/*
 * 	libMD407
 *  driver_usart.c
 *  Read buffered USART driver
 *  Drivers for STDIN, STDOUT and STDERR
 */
#include	"libmd407.h"

/* Declare functions here */
static int usart_init( int initval );
static void usart_deinit( int deinitval);
static int usart_write(char *ptr, int len);
static int usart_read(char *ptr, int len);
static int usart_isatty(void);

DEV_DRIVER_DESC StdIn = 
{
  {"stdin"},
  usart_init,
  usart_deinit,
  0,
  usart_isatty,
  0,
  0,
  0,
  0,
  usart_read
};

DEV_DRIVER_DESC StdOut = 
{
  {"stdout"},
  usart_init,
  usart_deinit,
  0,
  usart_isatty,
  0,
  0,
  0,
  usart_write,
  0
};
DEV_DRIVER_DESC StdErr = 
{
  {"stderr"},
  usart_init,
  usart_deinit,
  0,
  usart_isatty,
  0,
  0,
  0,
  usart_write,
  0
};

/* USART definitions */

typedef struct tag_usart
{
	volatile unsigned short sr;
	volatile unsigned short Unused0;
	volatile unsigned short dr;
	volatile unsigned short Unused1;
	volatile unsigned short brr;
	volatile unsigned short Unused2;
	volatile unsigned short cr1;
	volatile unsigned short Unused3;
	volatile unsigned short cr2;
	volatile unsigned short Unused4;
	volatile unsigned short cr3;
	volatile unsigned short Unused5;
	volatile unsigned short gtpr;
} USART;

#define	EN		(1<<13)
#define	TXEIE	(1<<7)
#define	TCIE	(1<<6)
#define	RXNEIE	(1<<5)
#define	TE		(1<<3)
#define	RE		(1<<2)

#define	TXE		(1<<7)	/* Transmit data register empty */
#define	RXNE	(1<<5)	/* Receive data register full */

#define USART1  ((USART *) 0x40011000)

#define	USART1_IRQVEC 	((void (**)(void) ) 0x2001C0D4 ) 

typedef	void (*irq_routine )(void) ; 


#define	NVIC_USART1_ISER 		0xE000E104  
#define NVIC_USART1_IRQ_BPOS  	(1<<5)



void EnableInterrupts( void ) __attribute__( ( naked ) ) ;
void EnableInterrupts( void )
{
	__asm (
	"		cpsie 	i\n"		/* set I=0 */
	"		bx		lr\n"
	);
}
void DisableInterrupts( void ) __attribute__( ( naked ) ) ;
void DisableInterrupts( void )
{
	__asm (
	"		cpsid 	i\n"		/* set I=1 */
	"		bx		lr\n"
	);
}

/* Buffering ... */
#define		USARTBUFFSIZE 32
typedef struct{
    unsigned char in;
    unsigned char out;
    unsigned char count;
    unsigned char buff[USARTBUFFSIZE];
}FIFO;

static FIFO	RxBuf;


static void bufferInit( FIFO *buffer )
{
	buffer->count = 0;		//0 bytes in buffer
	buffer->in = 0;			//index points to start
	buffer->out = 0;		//index points to start
}

static int bufferPut( FIFO *buffer, char ch )
{
	if( buffer->count==USARTBUFFSIZE ){
		return 0;		//buffer full
	}	
	DisableInterrupts();
	buffer->buff[buffer->in++]=ch;
	buffer->count++;
	if(buffer->in==USARTBUFFSIZE)
		buffer->in=0;  //wrap
	EnableInterrupts();
	return 1;
}

static int bufferGet( FIFO *buffer, char *ch)
{
	if(buffer->count==0)
		return 0;	//buffer empty
	DisableInterrupts();
	*ch=buffer->buff[buffer->out++];
	buffer->count--;
	if(buffer->out==USARTBUFFSIZE)
		buffer->out=0;//start from beginning
	EnableInterrupts();
	return 1;
}



void	spin(void)
{
	while(1);
}

static void	usart_service_routine( void )
{
	char	c;
	if( USART1->sr & RXNE)
	{ 
		if ( ! bufferPut( &RxBuf , (char) USART1->dr ) )
		{  spin();/* Buffer full ... */ }
	}		
}

static int usart_init( int initval )
{
    static int initted = 0;
    if ( initted ) return 1;
	*((unsigned int *) NVIC_USART1_ISER) |= NVIC_USART1_IRQ_BPOS;
	*((void (**)(void) ) 0x2001C0D4 )  = usart_service_routine;	
	/* Init buffers */
	bufferInit( &RxBuf );	
	
	/* Setup USART, receive with IRQ, polled transmit, 10 bit (1 startbit, 8 databits */ 
	USART1->brr = 0x2D9;	/* Baudrade = 115200 */
	USART1->cr3 = 0;
	USART1->cr2 = 0;	/* 1 stopbit */
#ifdef	USE_IRQ
	USART1->cr1 = EN | RXNEIE | TE | RE;
#else
	USART1->cr1 = EN | TE | RE;
#endif
    initted++;
    return 1; /* Ok init */
}


static void usart_deinit( int deinitval)
{
	fflush(stdout); /* Will cause terminal flush... */
    /* Disable USART and interrupts */
    USART1->cr1 = 0;
}

static int usart_isatty(void)
{
		return 1;
}

static char	usart_get ( void )
{
	char	c;
	if( bufferGet ( &RxBuf , &c ) )
		return c;
	return 0;
}


static char	_tstchar( void )
{
	if( (USART1->sr & (1<<5) )==0)
		return 0;
	return (char) USART1->dr;
}


static int usart_read(char *ptr, int len)
{
    char c;
	int items = 0;
    while( 1 )
    {
#ifdef	USE_IRQ
        if( bufferGet ( &RxBuf , &c ) )
# else
		c = _tstchar();
		if( c != 0)
#endif
        {
            *ptr++ = c;
             items++;
        }
		if( len == items)
			return items;
    }
}



static void	_outchar( char c )
{
	/* write character to usart1 */
 	while (( USART1->sr & 0x80)==0);
	USART1->dr = (unsigned short) c;
	if( c == '\n')
			_outchar('\r');
}


static int usart_write(char *ptr, int len)
{
    int rval = len;
 	while (len--)
      _outchar(*ptr++);
	return rval;
}





